/* mpz_bin_uiui - compute n over k.

Copyright 2011 Free Software Foundation, Inc.

This file is part of the GNU MP Library.

The GNU MP Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 3 of the License, or (at your
option) any later version.

The GNU MP Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with the GNU MP Library.  If not, see http://www.gnu.org/licenses/.  */

#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"

/* Algorithm:

   Accumulate chunks of factors first limb-by-limb (using one of mul0-mul8) and
   then into small mpn numbers.  The 1st inner loop accumulates dividend
   factors, the 2nd accumulates exactly the same number of divisor factors.  We
   avoid accumulating more for the divisor, even with its smaller factors,
   since we else cannot guarantee divisibility.

   Since we now each division will yield an integer, we compute it using Hensel
   norm: If the quotient is limited by 2^t, we compute A / B mod 2^t.

   An obvious improvement to this code would be to compute mod 2^t everywhere.
   Unfortunately, we cannot determine t beforehand, unless we invoke some
   appromixation, such as Stirling's formula.

   Other improvements: When dividing out factors-of-two, avoid copying down
   numbers.
*/

/* This threshold determines how large dividend to accumulate before we call
   bdiv.  Perhaps we should never call bdiv, and accumulate all we are told,
   since we are just basecase code anyway?  */
#define SOME_THRESHOLD 20

static mp_limb_t
mul0 (mp_ptr p, mp_size_t n, mp_limb_t m)
{
  return m;
}

static mp_limb_t
mul1 (mp_ptr p, mp_size_t n, mp_limb_t m)
{
  p[n] = mpn_mul_1 (p, p, n, m);
  return m + 1;
}

static mp_limb_t
mul2 (mp_ptr p, mp_size_t n, mp_limb_t m)
{
  mp_limb_t m01 = (m + 0) * (m + 1);
  p[n] = mpn_mul_1 (p, p, n, m01);
  return m + 2;
}

static mp_limb_t
mul3 (mp_ptr p, mp_size_t n, mp_limb_t m)
{
  mp_limb_t m01 = (m + 0) * (m + 1);
  mp_limb_t m2 = (m + 2);
  p[n] = mpn_mul_1 (p, p, n, m01 * m2);
  return m + 3;
}

static mp_limb_t
mul4 (mp_ptr p, mp_size_t n, mp_limb_t m)
{
  mp_limb_t m01 = (m + 0) * (m + 1);
  mp_limb_t m23 = (m + 2) * (m + 3);
  p[n] = mpn_mul_1 (p, p, n, m01 * m23);
  return m + 4;
}

static mp_limb_t
mul5 (mp_ptr p, mp_size_t n, mp_limb_t m)
{
  mp_limb_t m012 = (m + 0) * (m + 1) * (m + 2);
  mp_limb_t m34 = (m + 3) * (m + 4);
  p[n] = mpn_mul_1 (p, p, n, m012 * m34);
  return m + 5;
}

static mp_limb_t
mul6 (mp_ptr p, mp_size_t n, mp_limb_t m)
{
  mp_limb_t m01 = (m + 0) * (m + 1);
  mp_limb_t m23 = (m + 2) * (m + 3);
  mp_limb_t m45 = (m + 4) * (m + 5);
  mp_limb_t m0123 = m01 * m23;
  p[n] = mpn_mul_1 (p, p, n, m0123 * m45);
  return m + 6;
}

static mp_limb_t
mul7 (mp_ptr p, mp_size_t n, mp_limb_t m)
{
  mp_limb_t m01 = (m + 0) * (m + 1);
  mp_limb_t m23 = (m + 2) * (m + 3);
  mp_limb_t m456 = (m + 4) * (m + 5) * (m + 6);
  mp_limb_t m0123 = m01 * m23;
  p[n] = mpn_mul_1 (p, p, n, m0123 * m456);
  return m + 7;
}

static mp_limb_t
mul8 (mp_ptr p, mp_size_t n, mp_limb_t m)
{
  mp_limb_t m01 = (m + 0) * (m + 1);
  mp_limb_t m23 = (m + 2) * (m + 3);
  mp_limb_t m45 = (m + 4) * (m + 5);
  mp_limb_t m67 = (m + 6) * (m + 7);
  mp_limb_t m0123 = m01 * m23;
  mp_limb_t m4567 = m45 * m67;
  p[n] = mpn_mul_1 (p, p, n, m0123 * m4567);
  return m + 8;
}

typedef mp_limb_t (* mulfunc_t) (mp_ptr, mp_size_t, mp_limb_t);

static mulfunc_t mulfunc[] = {mul0,mul1,mul2,mul3,mul4,mul5,mul6,mul7,mul8};
#define M (numberof(mulfunc)-1)

/* Rough computation of how many factors we can multiply together
   without spilling a limb.  */
#define MAXFACS(max,l)							\
  do {									\
    int __cnt, __logb;							\
    count_leading_zeros (__cnt, (mp_limb_t) (l));			\
    __logb = GMP_LIMB_BITS - __cnt;					\
    (max) = GMP_NUMB_BITS / __logb;	/* FIXME: remove division */	\
  } while (0)

void
mpz_bin_uiui (mpz_ptr r, unsigned long int n, unsigned long int k)
{
  int nmax, kmax, kmaxnow, nnumfac;
  mulfunc_t nfn, kfn;
  mp_ptr np, kp, qp;
  mp_size_t nn, kn, qn;
  mp_limb_t i, j, t;
  mp_limb_t dinv;
  int cnt;
  TMP_DECL;
  TMP_MARK;

  np = TMP_ALLOC_LIMBS (10000);
  kp = TMP_ALLOC_LIMBS (10000);
  qp = TMP_ALLOC_LIMBS (10000);

  MAXFACS (nmax, n);
  nmax = MIN (nmax, M);
  MAXFACS (kmax, k);
  kmax = MIN (kmax, M);

  j = 1;
  i = n - k + 1;

  t = n - i + 1;
  nmax = MIN (nmax, t);
  t = k - j + 1;
  kmax = MIN (kmax, t);

  np[0] = 1; nn = 1;

  while (i <= n || j <= k)
    {
      kp[0] = 1; kn = 1;

      nnumfac = i;
      while (nmax != 0 && nn <  SOME_THRESHOLD)
	{
	  nfn = mulfunc[nmax];

	  i = nfn (np, nn, i);
	  nn += np[nn] != 0;
	  if (np[0] == 0)
	    {
	      MPN_COPY_INCR (np, np + 1, nn - 1);
	      nn--;
	    }
	  t = n - i + 1;
	  nmax = MIN (nmax, t);
	}
      nnumfac = i - nnumfac;

      while (nnumfac != 0)
	{
	  kmaxnow = MIN (kmax, nnumfac);
	  kfn = mulfunc[kmaxnow];
	  nnumfac -= kmaxnow;

	  j = kfn (kp, kn, j);
	  kn += kp[kn] != 0;
	  if (kp[0] == 0)
	    {
	      MPN_COPY_INCR (kp, kp + 1, kn - 1);
	      kn--;
	    }
	}

      count_trailing_zeros (cnt, kp[0]);
      if (cnt != 0)
	{
	  if ((np[0] << (GMP_NUMB_BITS - cnt)) != 0)
	    {
	      /* Ugly and not obviously right.  This happens when we truncated
		 more zero limbs from np[] than from kp[].  */
	      np[nn] = mpn_lshift (np, np, nn, GMP_NUMB_BITS - cnt);
	      nn += np[nn] != 0;
	    }
	  else
	    {
	      mpn_rshift (np, np, nn, cnt);
	      nn -= np[nn - 1] == 0;
	    }
	  mpn_rshift (kp, kp, kn, cnt);
	  kn -= kp[kn - 1] == 0;
	}

      binvert_limb (dinv, kp[0]);
      if (nn == kn)
	{
	  np[0] = dinv * np[0];
	  nn = 1;
	}
      else
	{
	  if (mpn_cmp (np + nn - kn, kp, kn) > 0)
	    np[nn++] = 0;
	  qn = nn - kn;
	  mpn_sbpi1_bdiv_q (qp, np, qn, kp, MIN(kn,qn), -dinv);
	  nn = qn;
	  MPN_COPY (np, qp, nn);
	}
    }

  nn -= np[nn - 1] == 0;
  MPZ_REALLOC (r, nn);
  SIZ(r) = nn;
  MPN_COPY (PTR(r), np, nn);
  TMP_FREE;
}

#include <stdlib.h>

int
main (int argc, char **argv)
{
  unsigned long n, k;
  mpz_t r;

  n = strtoul (argv[1], 0, 0);
  k = strtoul (argv[2], 0, 0);

  mpz_init (r);

  mpz_bin_uiui (r, n, k);
  gmp_printf ("%Zx\n", r);

  return 0;
}
